/* Program: ex3-1.c
 * ---------------------------------------------------------------------
 * Solution to exercise 1, series 3.
 *
 * This program transforms a 2-dimensional matrix of special form
 * (like a checker board) to a 1-dimensional C array.
 * ---------------------------------------------------------------------
 * Nikolaos S. Papaspyrou (nickie@softlab.ntua.gr)
 * National Technical University of Athens
 * Department of Electrical Engineering
 * ---------------------------------------------------------------------
 * Course: Programming Techniques (2nd semester)
 * URL:    http://www.softlab.ntua.gr/~nickie/Courses/progtech/
 * ---------------------------------------------------------------------
 */


/* Header files */

#include <stdio.h>


/*
   The number of non-zero squares in the 2-dimensional matrix is:

      (N/2)*N              N for every two lines
    + (N%2)*((N+1)/2)      N/2 rounded up to the next integer
                              but only if N is odd

   For N=2*K, this gives: 2*K^2
   For N=2*K+1, this gives: K*(2*K+1) + K+1 = 2*K^2 + 2*K+1

   In other words, the number is:

      (N^2)/2                if N is even
      2 * (N div 2)^2 + N    if N is odd

   In our example program, for N=8 we get 32 and for N=9 we get 41
 */


/* Function that computes the index of a matrix element in the array.
   Assuming that 0 <= i,j < n.
 */

int loc (int n, int i, int j)
{
   int m = (n+1) / 2;      /* this is n/2 rounded up to the next integer */

   if ((i+j) % 2 == 0)     /* if it is one of the non-zero squares ...   */
      return (i/2) * n     /*    every two lines, we add n elements      */
           + (i%2) * m     /*    plus m elements on the previous line,   */
                           /*       if we haven't counted it,            */
           + j/2;          /*    plus j/2 elements on the same line      */
   else                    /* else return something invalid              */
      return -1;
}


/* Unfortunately, in the conversion functions we need to define an
   upper limit for the (second) dimension of the matrix; we will only
   use a submatrix of dimension n x n inside the big one...
 */

#define MAX 20


/* Function that translates u -> c */

void utoc (int n, int u[][MAX], int c[])
{
   int i=0, j=0, k=0;

   while (i < n) {
      c[k++] = u[i][j];
      if ((j += 2) >= n)   /* increase j by 2; if it's out of the matrix */
         j = (++i) % 2;    /*    increase i and set j to the right value */
   }
}


/* Function that translates c -> u */

void ctou (int n, int c[], int u[][MAX])
{
   int i, j, k=0;

   for (i=0; i<n; i++)
      for (j=0; j<n; j++)
         if ((i+j) % 2 == 0)     /* if it is one of the non-zero squares */
            u[i][j] = c[k++];
         else
            u[i][j] = 0;
}


/* Main program */

int main ()
{
   int u8[8][MAX] = { {  1,  0,  2,  0,  3,  0,  4,  0 },
                      {  0,  5,  0,  6,  0,  7,  0,  8 },
                      {  9,  0, 10,  0, 11,  0, 12,  0 },
                      {  0, 13,  0, 14,  0, 15,  0, 16 },
                      { 17,  0, 18,  0, 19,  0, 20,  0 },
                      {  0, 21,  0, 22,  0, 23,  0, 24 },
                      { 25,  0, 26,  0, 27,  0, 28,  0 },
                      {  0, 29,  0, 30,  0, 31,  0, 32 }
                    };

   int u9[9][MAX] = { {  1,  0,  2,  0,  3,  0,  4,  0,  5 },
                      {  0,  6,  0,  7,  0,  8,  0,  9,  0 },
                      { 10,  0, 11,  0, 12,  0, 13,  0, 14 },
                      {  0, 15,  0, 16,  0, 17,  0, 18,  0 },
                      { 19,  0, 20,  0, 21,  0, 22,  0, 23 },
                      {  0, 24,  0, 25,  0, 26,  0, 27,  0 },
                      { 28,  0, 29,  0, 30,  0, 31,  0, 32 },
                      {  0, 33,  0, 34,  0, 35,  0, 36,  0 },
                      { 37,  0, 38,  0, 39,  0, 40,  0, 41 }
                    };

   int c8[32], c9[45];
   int w8[8][MAX], w9[9][MAX];
   
   int i, j;

   printf("\nFor N=8\n\n");
   utoc(8, u8, c8);
   for (i=0; i<8; i++) {
      printf("   ");
      for (j=0; j<8; j++) {
         int k = loc(8, i, j);

         if (k >= 0)
            printf("%3d", c8[k]);
         else
            printf("   ");
      }
      printf("\n");
   }
   ctou(8, c8, w8);
   for (i=0; i<8; i++)
      for (j=0; j<8; j++)
         if (u8[i][j] != w8[i][j]) {
            printf("Something wrong with w8[%d][%d]\n", i, j);
            return 1;
         }

   printf("\nFor N=9\n\n");
   utoc(9, u9, c9);
   for (i=0; i<9; i++) {
      printf("   ");
      for (j=0; j<9; j++) {
         int k = loc(9, i, j);

         if (k >= 0)
            printf("%3d", c9[k]);
         else
            printf("   ");
      }
      printf("\n");
   }
   ctou(9, c9, w9);
   for (i=0; i<9; i++)
      for (j=0; j<9; j++)
         if (u9[i][j] != w9[i][j]) {
            printf("Something wrong with w9[%d][%d]\n", i, j);
            return 1;
         }

   return 0;
}
