/* Program: ex3-2.c
 * ---------------------------------------------------------------------
 * Solution to exercise 2, series 3.
 *
 * This program converts a sparse matrix to a linked list and back.
 * ---------------------------------------------------------------------
 * Nikolaos S. Papaspyrou (nickie@softlab.ntua.gr)
 * National Technical University of Athens
 * Department of Electrical Engineering
 * ---------------------------------------------------------------------
 * Course: Programming Techniques (2nd semester)
 * URL:    http://www.softlab.ntua.gr/~nickie/Courses/progtech/
 * ---------------------------------------------------------------------
 */


/* Header files */

#include <stdio.h>
#include <stdlib.h>


/* Definition of the linked list */

typedef struct node_tag {
   int row, col;
   int data;
   struct node_tag * next;
} Node, *List;


/* Unfortunately, in the conversion functions we need to define an
   upper limit for the (second) dimension of the matrix; we will only
   use a submatrix of dimension n x m inside the big one...
 */

#define MAX 20


/* Conversion from matrix to list */

List mtol (int n, int m, int a[][MAX])
{
   List l = NULL, t;
   int i, j;

   /* Traverse the matrix in reverse order, so we can insert elements
      at the head of the list */

   for (i=n-1; i>=0; i--)
      for (j=m-1; j>=0; j--)
         if (a[i][j] != 0) {
            if ((t = (Node *) malloc(sizeof(Node))) == NULL) {
               fprintf(stderr, "Out of memory\n");
               exit(1);
            }

            t->row = i;
            t->col = j;
            t->data = a[i][j];
            t->next = l;
            l = t;
         }

   /* Add one element with the dimensions of the matrix */

   if ((t = (Node *) malloc(sizeof(Node))) == NULL) {
      fprintf(stderr, "Out of memory\n");
      exit(1);
   }

   t->row = n;
   t->col = m;
   t->data = 0;
   t->next = l;
   return t;
}


/* Conversion from list to matrix */

void ltom (List l, int a[][MAX])
{
   List t;
   int n, m, i, j;

   if (l == NULL) {
      fprintf(stderr, "The first element of the list is missing\n");
      exit(1);
   }

   n = l->row;
   m = l->col;

   for (i=0; i<n; i++)
      for (j=0; j<m; j++)
         a[i][j] = 0;

   for (t = l->next; t != NULL; t = t->next)
      a[t->row][t->col] = t->data;
}


/* Main program */

int main ()
{
   int a[MAX][MAX], b[MAX][MAX];
   int n, m, i, j;
   List l, t;

   printf("Give the number of rows,    n = ");
   scanf("%d", &n);
   printf("Give the number of columns, m = ");
   scanf("%d", &m);

   if (n > MAX || m > MAX) {
      fprintf(stderr, "I can only manage up to %dx%d matrices\n", MAX, MAX);
      return 1;
   }

   printf("Give all the %dx%d elements of the matrix:\n", n, m);
   for (i=0; i<n; i++)
      for (j=0; j<m; j++)
         scanf("%d", &(a[i][j]));

   printf("\nHere is your matrix:\n\n");
   for (i=0; i<n; i++) {
      printf("   ");
      for (j=0; j<m; j++)
         printf("%3d", a[i][j]);
      printf("\n");
   }

   l = mtol(n, m, a);

   printf("\nHere is the list:\n\n");
   for (t = l; t != NULL; t = t->next) {
      printf("(%d, %d, %d)", t->row, t->col, t->data);
      if (t->next != NULL)
         printf(" -> ");
   }
   printf("\n");

   ltom(l, b);

   printf("\nHere is the reconstructed matrix:\n\n");
   for (i=0; i<n; i++) {
      printf("   ");
      for (j=0; j<m; j++)
         printf("%3d", b[i][j]);
      printf("\n");
   }

   for (i=0; i<n; i++)
      for (j=0; j<m; j++)
         if (b[i][j] != a[i][j]) {
            printf("Something wrong with b[%d][%d]\n", i, j);
            return 1;
         }

   return 0;
}
