/* Program: ex3-3.c
 * ---------------------------------------------------------------------
 * Solution to exercise 3, series 3.
 *
 * This program represents polynoms as linked lists.
 * ---------------------------------------------------------------------
 * Nikolaos S. Papaspyrou (nickie@softlab.ntua.gr)
 * National Technical University of Athens
 * Department of Electrical Engineering
 * ---------------------------------------------------------------------
 * Course: Programming Techniques (2nd semester)
 * URL:    http://www.softlab.ntua.gr/~nickie/Courses/progtech/
 * ---------------------------------------------------------------------
 */


/* Header files */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>


/* Definition of a polynom as a linked list.
   We assume that polynoms are sorted in ascending order of the exponent
   and that coefficients are not zero.
 */

typedef struct node_tag {
   int exponent;
   double coefficient;
   struct node_tag * next;
} Node, *Polynom;


/* Print a polynom */

void polynomPrint (Polynom p)
{
   if (p != NULL) {
      double coeff = p->coefficient;
      int expon = p->exponent;

      if (p->next != NULL)
         polynomPrint(p->next);
      if (p->next != NULL && coeff > 0.0)
         printf(" + ");
      else if (coeff < 0.0) {
         printf(" - ");
         coeff = -coeff;
      }
      if (coeff != 1.0 || expon == 0)
         printf("%lg", coeff);
      if (expon > 0) {
         printf("x");
         if (expon > 1)
            printf("^%d", expon);
      }
   }
   else
      printf("0");
}


/* Insert (prepend) a term to a polynom.
   This function could make sure that terms are in ascending order,
   but it does not, as we have a special function for this purpose.
 */

Polynom polynomInsert (double coeff, int expon, Polynom p)
{
   if (coeff != 0.0) {
      Node * n;

      if ((n = (Node *) malloc(sizeof(Node))) == NULL) {
         fprintf(stderr, "Out of memory\n");
         return NULL;
      }

      n->coefficient = coeff;
      n->exponent = expon;
      n->next = p;
      
      return n;
   }
   else
      return p;
}


/* Read a polynom of variable v.
   This is more than the exercise requires...
 */

#define MAX 32

Polynom polynomRead (char v)
{
   Polynom p = NULL;
   int c = getchar();

   /* Skip all spaces and pending newlines */
   
   while (isspace(c))
      c = getchar();

   /* For all terms */
   
   while (c != EOF && c != '\n') {
      double coeff;
      int expon;
      char sign;
      char buffer[MAX];
      int i;
      int term = 0;

      /* Skip spaces */

      if (c == ' ') {
         c = getchar();
         continue;
      }

      /* Read sign */

      if (c == '+' || c == '-') {
         sign = c;
         c = getchar();

         /* Skip spaces */

         while (c == ' ')
            c = getchar();
      }
      else
         sign = '+';

      /* Read coefficient */

      i = 0;
      while (isdigit(c) || c == '.') {
         buffer[i++] = c;
         c = getchar();
      }
      if (i > 0) {
         buffer[i] = '\0';
         coeff = atof(buffer);
         term = 1;

         /* Skip spaces */

         while (c == ' ')
            c = getchar();
      }
      else
         coeff = 1.0;

      if (sign == '-')
         coeff = -coeff;

      /* Read variable */

      if (c == v) {
         c = getchar();
         term = 1;

         /* Skip spaces */

         while (c == ' ')
            c = getchar();

         /* Read ^ character */

         if (c == '^') {
            c = getchar();

            /* Skip spaces */

            while (c == ' ')
               c = getchar();

            /* Read exponent */

            i = 0;
            while (isdigit(c)) {
               buffer[i++] = c;
               c = getchar();
            }
            if (i > 0) {
               buffer[i] = '\0';
               expon = atoi(buffer);
            }
            else {
               fprintf(stderr, "Exponent is missing...\n");
               return NULL;
            }
         }

         /* If no ^ character, first order term */

         else
            expon = 1;
      }

      /* If no variable, zero order term */

      else
         expon = 0;

      /* If we have a term, add it ... */
      
      if (term)
         p = polynomInsert(coeff, expon, p);
      else {
         fprintf(stderr, "Invalid term\n");
         return NULL;
      }
   }

   return p;
}


/* Check if a polynom is valid */

int polynomValid (Polynom p)
{
   int lastExpon = -1;
   
   while (p != NULL) {
      if (p->coefficient == 0.0) {
         fprintf(stderr, "Zero coefficients should not be present\n");
         return 0;
      }
      if (p->exponent <= lastExpon) {
         fprintf(stderr, "Exponents are not in ascending order\n");
         return 0;
      }
      lastExpon = p->exponent;
      p = p->next;
   }
   
   return 1;      
}


/* Evaluate a polynom */

double polynomEvaluate (Polynom p, double x)
{
   double result    = 0.0;
   double lastPower = 1.0;
   int    lastExpon = 0;

   while (p != NULL) {

      /* Take advantage of the ascending order of exponents to
         calculate only the "difference" factor */

      int diffExpon = p->exponent - lastExpon;
      int power = x;

      /* Play with the bits of the exponent to calculate the
         power efficiently */

      while (diffExpon > 0) {
         if (diffExpon & 1)
            lastPower *= power;
         power *= power;
         diffExpon >>= 1;
      }

      result += p->coefficient * lastPower;
      lastExpon = p->exponent;
      p = p->next;
   }

   return result;
}


/* Add two polynoms.
   This function is recursive.  It takes advantage of the fact that
   polynoms do not change over time, so the lists it creates may reuse
   parts of the two operands.
*/ 

Polynom polynomAdd (Polynom p, Polynom q)
{
   if (p == NULL)
      return q;
   else if (q == NULL)
      return p;
   else if (p->exponent < q->exponent)
      return polynomInsert(p->coefficient, p->exponent,
                           polynomAdd(p->next, q));
   else if (q->exponent < p->exponent)
      return polynomInsert(q->coefficient, q->exponent,
                           polynomAdd(p, q->next));
   else
      return polynomInsert(p->coefficient + q->coefficient, p->exponent,
                           polynomAdd(p->next, q->next));
}


/* This is an auxiliary function used in polynom multiplication.
   It calculates the polynom:  a x^n p(x) + r(x)
 */
 
Polynom polynomMultiAdd (double a, int n, Polynom p, Polynom r)
{
   if (p == NULL)
      return r;
   else if (r == NULL || n + p->exponent < r->exponent)
      return polynomInsert(a * p->coefficient,
                           n + p->exponent,
                           polynomMultiAdd(a, n, p->next, r));
   else if (r->exponent < n + p->exponent)
      return polynomInsert(r->coefficient,
                           r->exponent,
                           polynomMultiAdd(a, n, p, r->next));
   else
      return polynomInsert(a * p->coefficient + r->coefficient,
                           r->exponent,
                           polynomMultiAdd(a, n, p->next, r->next));
}

 
/* Multiply two polynoms.
   This function is recursive.  It takes advantage of the fact that
   polynoms do not change over time, so the lists it creates may reuse
   parts of the two operands.
*/

Polynom polynomMultiply (Polynom p, Polynom q)
{
   if (p == NULL)
      return NULL;
   else if (q == NULL)
      return NULL;
   else {
      Polynom r1 = polynomMultiply(p->next, q->next);
      Polynom r2 = polynomMultiAdd(p->coefficient, p->exponent, q->next, r1);
      Polynom r3 = polynomMultiAdd(q->coefficient, q->exponent, p->next, r2);
      
      return polynomInsert(p->coefficient * q->coefficient,
                           p->exponent + q->exponent, r3);
   }
}


/* Main program */

int main ()
{
   Polynom p, q, r1, r2;
   double x;

   printf("Give a polynom P(x): ");
   p = polynomRead('x');
   
   if (!polynomValid(p))
      return 1;

   printf("P(x) = ");
   polynomPrint(p);
   printf("\n");

   printf("Give a value for x: ");
   scanf("%lf", &x);
   printf("P(%lg) = %lg\n", x, polynomEvaluate(p, x));

   printf("Give a polynom Q(x): ");
   q = polynomRead('x');
   
   if (!polynomValid(q))
      return 1;

   r1 = polynomAdd(p, q);
   
   printf("(P+Q)(x) = ");
   polynomPrint(r1);
   printf("\n");
   
   if (!polynomValid(r1))
      return 1;

   r2 = polynomMultiply(p, q);
   
   printf("(P*Q)(x) = ");
   polynomPrint(r2);
   printf("\n");
   
   if (!polynomValid(r2))
      return 1;

   return 0;
}
