/* Program: ex3-6.c
 * ---------------------------------------------------------------------
 * Solution to exercise 6, series 3.
 *
 * This program flattens nested linked lists.
 * ---------------------------------------------------------------------
 * Nikolaos S. Papaspyrou (nickie@softlab.ntua.gr)
 * National Technical University of Athens
 * Department of Electrical Engineering
 * ---------------------------------------------------------------------
 * Course: Programming Techniques (2nd semester)
 * URL:    http://www.softlab.ntua.gr/~nickie/Courses/progtech/
 * ---------------------------------------------------------------------
 */


/* Header files */

#include <stdio.h>
#include <stdlib.h>


/* Definition of a nested linked list */

typedef struct node_tag {
   enum { SIMPLE, LIST } type;
   union {
      int integer;
      struct node_tag * list;
   } data;
   struct node_tag * next;
} Node, *NestedList;


/* Insert (prepend) an integer to a nested linked list */

NestedList listInsertInteger (int d, NestedList l)
{
   Node * n = (Node *) malloc(sizeof(Node));
   
   if (n == NULL) {
      fprintf(stderr, "Out of memory\n");
      exit(1);
   }

   n->type = SIMPLE;
   n->data.integer = d;
   n->next = l;
   return n;
}


/* Insert (prepend) a list to a nested linked list */

NestedList listInsertList (NestedList d, NestedList l)
{
   Node * n = (Node *) malloc(sizeof(Node));
   
   if (n == NULL) {
      fprintf(stderr, "Out of memory\n");
      exit(1);
   }

   n->type = LIST;
   n->data.list = d;
   n->next = l;
   return n;
}


/* Print a nested linked list */

void listPrint (NestedList l)
{
   printf("[");
   while (l != NULL) {
      switch (l->type) {
         case SIMPLE:
            printf("%d", l->data.integer);
            break;
         case LIST:
            listPrint(l->data.list);
            break;
      }
      if (l->next != NULL)
         printf(", ");
      l = l->next;
   }   
   printf("]");
}


/* Auxiliary function.
   Create a new list by flattening a given nested linked list.
   It returns two pointers: one to the first and one to the last node
   in the flat list.  We need these two to speed up the appending of
   two lists.
 */

void flattenAuxiliary (NestedList l, NestedList * lfirst, NestedList * llast)
{
   if (l == NULL)
      *lfirst = *llast = NULL;
   else {
      NestedList ltfirst, ltlast;
      
      flattenAuxiliary(l->next, &ltfirst, llast);
      switch (l->type) {
         case SIMPLE:
            *lfirst = listInsertInteger(l->data.integer, ltfirst);;
            if (*llast == NULL)
               *llast = *lfirst;
            break;
         case LIST:
            flattenAuxiliary(l->data.list, lfirst, &ltlast);
            if (ltlast == NULL)
               *lfirst = ltfirst;
            else
               ltlast->next = ltfirst;
            if (ltfirst == NULL)
               *llast = ltlast;
            break;
      }
   }
}


/* Create a new list by flattening a given nested linked list */

NestedList flatten (NestedList l)
{
   NestedList lfirst, llast;
   
   flattenAuxiliary(l, &lfirst, &llast);
   return lfirst;
}


/* Main program */

int main ()
{
   NestedList l =
      listInsertInteger(
         7,
         listInsertList(
            listInsertInteger(
               1,
               listInsertInteger(
                  42,
                  listInsertInteger(
                     3,
                     NULL
                  )
               )
            ),
            listInsertInteger(
               6,
               listInsertList(
                  listInsertInteger(
                     12,
                     listInsertList(
                        listInsertInteger(
                           2,
                           listInsertInteger(
                              7,
                              NULL
                           )
                        ),
                        listInsertInteger(
                           14,
                           NULL
                        )
                     )
                  ),
                  NULL
               )
            )
         )
      );
   
   printf("The original list is: ");
   listPrint(l);
   printf("\n");

   printf("The flattened list is: ");
   listPrint(flatten(l));
   printf("\n");

   return 0;
}
