/* Program: ex3-7.c
 * ---------------------------------------------------------------------
 * Solution to exercise 7, series 3.
 *
 * This program reads a sequence of integers from INFILE and prints it
 * to OUTFILE by alternating elements from the start end from the end.
 * Using a doubly linked list.
 * ---------------------------------------------------------------------
 * Nikolaos S. Papaspyrou (nickie@softlab.ntua.gr)
 * National Technical University of Athens
 * Department of Electrical Engineering
 * ---------------------------------------------------------------------
 * Course: Programming Techniques (2nd semester)
 * URL:    http://www.softlab.ntua.gr/~nickie/Courses/progtech/
 * ---------------------------------------------------------------------
 */


/* Header files */

#include <stdio.h>
#include <stdlib.h>


/* Doubly linked lists */

typedef struct DListNode_tag {
   int data;
   struct DListNode_tag * next;
   struct DListNode_tag * prev;
} DListNode;


/* Type dlist */

typedef struct {
   DListNode * first;
   DListNode * last;
} dlist;


/* Empty list */

const dlist dlistEmpty = { NULL, NULL };


/* Check if a list is empty */

int dlistIsEmpty (dlist l)
{
   return l.first == NULL;
}


/* Insert as last */

void dlistInsertLast (dlist * lp, int t)
{
   DListNode * n = (DListNode *) malloc(sizeof(DListNode));

   /* Check if malloc succeeded */
      
   if (n == NULL) {
      fprintf(stderr, "Out of memory\n");
      exit(1);
   }

   /* Copy the data */

   n->data = t;

   /* If the list was empty, insert just this element */ 

   if (lp->first == NULL) {
      n->prev = n->next = NULL;
      lp->first = lp->last = n;
   }

   /* Else, insert as last element */
   
   else {
      n->prev = lp->last;
      n->next = NULL;
      lp->last->next = n;
      lp->last = n;
   }
}


/* Remove first */

int dlistRemoveFirst (dlist * lp)
{
   int result;
   
   /* Error, if the list was empty */
   
   if (lp->first == NULL) {
      fprintf(stderr, "Nothing to remove from empty list\n");
      exit(1);
   }

   result = lp->first->data;

   /* If just one element, remove this */
   
   if (lp->first == lp->last) {
      free(lp->first);
      lp->first = lp->last = NULL;
   }
         
   /* Otherwise, remove the first element */

   else {
      lp->first = lp->first->next;
      free(lp->first->prev);
      lp->first->prev = NULL;
   }
   
   return result; 
}


/* Remove last */

int dlistRemoveLast (dlist * lp)
{
   int result;
   
   /* Error, if the list was empty */
   
   if (lp->first == NULL) {
      fprintf(stderr, "Nothing to remove from empty list\n");
      exit(1);
   }

   result = lp->last->data;

   /* If just one element, remove this */
   
   if (lp->first == lp->last) {
      free(lp->first);
      lp->first = lp->last = NULL;
   }
         
   /* Otherwise, remove the first element */

   else {
      lp->last = lp->last->prev;
      free(lp->last->next);
      lp->last->next = NULL;
   }
   
   return result; 
}


/* Main program */

int main ()
{
   FILE * fin, * fout;
   dlist l = dlistEmpty;
   enum { START, END } whereFrom;
   int n;

   if ((fin = fopen("INFILE", "rt")) == NULL) {
      fprintf(stderr, "Cannot open INFILE\n");
      return 1;
   }
   if ((fout = fopen("OUTFILE", "wt")) == NULL) {
      fprintf(stderr, "Cannot open OUTFILE\n");
      return 1;
   }

   while (!feof(fin))
      if (fscanf(fin, "%d", &n) == 1)
         dlistInsertLast(&l, n);

   whereFrom = START;
   while (!dlistIsEmpty(l))
      switch (whereFrom) {
         case START:
            fprintf(fout, "%d\n", dlistRemoveFirst(&l));
            whereFrom = END;
            break;
         case END:
            fprintf(fout, "%d\n", dlistRemoveLast(&l));
            whereFrom = START;
            break;
      }

   return 0;
}
